/* Wine Profiler
 *
 * Copyright 2000-2003 Codeweavers, Charles Loep
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * It's unlikely that this works on anything other than linux/i386
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/resource.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "atomicity.h"
#include "profiler.h"


#define WPROF_MAX_CALL_DEPTH 1024
#define WPROF_MAX_ARC        1000000

#define WPROF_CURRENT        getpid()
#define WPROF_WMON_PREFIX    "/tmp/wmon.out."


#define DEBUG

#ifdef DEBUG
#define db_printf(args...) printf(args)
#else
#define db_printf(args...)
#endif


typedef struct cstack_s {
    arc_t      *current;
    counter64_t entry_time;
} cstack_t;


typedef struct ptinfo_s {
    pid_t pid;
    arc_t *root;
    int map_fd;
    int map_size;
    wprof_header_t *header;
    cstack_t cstack[WPROF_MAX_CALL_DEPTH];
    int cstackp;
    long state;
    struct ptinfo_s *next;
} ptinfo_t;

// Ok takes up some space, but at least we wont have to worry about locking
#define PTINFO_FINALIZED ((void *)-1)
#define PTINFO_TOTAL 32768
static ptinfo_t *ptinfo_map[PTINFO_TOTAL];


typedef enum {
    WPROF_STATE_OFF = 0,
    WPROF_STATE_ON,
    WPROF_STATE_BUSY
} WPROF_STATE;


/* global data */
static pid_t wprof_master = 0;
static counter32_t wprof_cpu_hz = 0;
static long wprof_global_state;


void __attribute__ ((constructor)) wprof_init();
void __attribute__ ((destructor)) wprof_exit();


static void wprof_finalize_thread(ptinfo_t *pt);
static arc_t *wprof_alloc_arc(ptinfo_t *pt, address32_t addr);


/*
 * determine the number of ticks per second
 */
static unsigned long wprof_get_hz()
{
    FILE *fp;
    double ticks = 0.0;

    fp = fopen("/proc/cpuinfo", "r");
    while(!feof(fp))
        {

            char line[1024];

            if(!fgets(line, 1024, fp))
                break;
            if(sscanf(line, "cpu MHz         : %lf", &ticks) != 1)
                continue;
            db_printf("WPROF: ticks per second: %.0lf\n", (ticks * 1000000.0));
            break;
        }
    fclose(fp);

    if(ticks == 0.0)
	{
	    printf("WPROF: unable to determine HZ (%lf)\n", ticks);
	    _exit(1);
	}

    return (unsigned long)(ticks * 1000000.0);
}

static counter64_t get_tsc() 
{
        counter64_t value;
        __asm__ __volatile__("rdtsc" : "=A" (value));
	return value;
}



/*
 * constructor
 */
void wprof_init()
{
    wprof_master = WPROF_CURRENT;
    wprof_cpu_hz = wprof_get_hz();
 
    db_printf("WPROF: initialized!\n");

    wprof_global_state = WPROF_STATE_ON;
}

/*
 * destructor
 */
void wprof_exit()
{
    int i;

    wprof_global_state = WPROF_STATE_OFF;
    printf("WPROF: unloading %d\n", WPROF_CURRENT);

#if 1
    // finalize any threads that didnt we didn't catch before
    for(i = 0; i < PTINFO_TOTAL; i++) {
	if(ptinfo_map[i]) {
	    ptinfo_t *pt = ptinfo_map[i];

	    if(pt == PTINFO_FINALIZED)
		    continue;

	    printf("WPROF: finalizing %d\n", pt->pid);
	    wprof_finalize_thread(pt);
	    ptinfo_map[i] = 0;
	}
    }
#endif

}


static char digits[] = { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9' };
static void strintcat(char *str, int i)
{
    int pos = 11;
    char buf[12];
        
    buf[11] = 0;
    do {
	buf[--pos] = digits[i % 10];
	i = i / 10;
    } while(i != 0);
    
    strcat(str, &buf[pos]);
}


/*
 * map the profile output file for the current thread
 */
static ptinfo_t *wprof_map_output()
{
    ptinfo_t *pt = NULL;
    char filename[256] = WPROF_WMON_PREFIX;
    void *addr;
    int map_size, map_fd;

    map_size = sizeof(wprof_header_t) + (WPROF_MAX_ARC * sizeof(arc_t)) + sizeof(ptinfo_t);

    strintcat(filename, WPROF_CURRENT);
    db_printf("WPROF: creating %s\n", filename);

    map_fd = open(filename, O_RDWR|O_CREAT|O_TRUNC, 0664);
    ftruncate(map_fd, map_size);
    addr = mmap(NULL, map_size, PROT_READ|PROT_WRITE, MAP_SHARED, map_fd, 0);

    pt = (ptinfo_t *)(addr + map_size - sizeof(ptinfo_t));
    pt->map_size = map_size;
    pt->map_fd = map_fd;
    pt->header = (wprof_header_t *)addr;
    pt->header->vaddr = (unsigned int)addr;

    return pt;
}



/* 
 * allocate a thread info struct
 */
static ptinfo_t *wprof_alloc_ptinfo()
{
    ptinfo_t *pt;
    arc_t *arc;
    pid_t current = WPROF_CURRENT;

    db_printf("WPROF: new thread; pid %d\n", current);

    pt = wprof_map_output();
    pt->pid  = current;
    pt->root = NULL;
    pt->next = NULL;

    pt->header->magic   = WPROF_MAGIC;
    pt->header->version = WPROF_VERSION;
    pt->header->pid     = pt->pid;
    pt->header->cpu_hz  = wprof_cpu_hz;
    gettimeofday(&pt->header->time_start, NULL);
    gettimeofday(&pt->header->time_stop, NULL);

    arc = wprof_alloc_arc(pt, 0);
    pt->root = arc;
    pt->cstackp = 0;
    pt->cstack[pt->cstackp].current = pt->root;
    pt->cstack[pt->cstackp].entry_time = get_tsc();

    ptinfo_map[current] = pt;

    //db_printf("WPROF: allocated ptinfo for pid %d\n", pt->pid);
    pt->state = WPROF_STATE_ON;

    return pt;
}


/*
 * retrieve or allocate a thread info struct
 */
static ptinfo_t *wprof_get_ptinfo()
{
    ptinfo_t *pt;
    pid_t current = WPROF_CURRENT;

    pt = ptinfo_map[current];
    
    if(pt == PTINFO_FINALIZED)
	    return NULL;

    if(!pt)
	    pt = wprof_alloc_ptinfo();
    return pt;
}


static void wprof_unlink_ptinfo(ptinfo_t *ptrem)
{
	ptinfo_map[ptrem->pid] = PTINFO_FINALIZED;
}

static int wprof_has_ptinfo()
{
    pid_t current = WPROF_CURRENT;

    if(ptinfo_map[current])
	return 1;
    return 0;
}

/*
 * allocate new arc
 * FIXME: make sure there's enough free space left for the allocation
 */
static arc_t *wprof_alloc_arc(ptinfo_t *pt, address32_t addr)
{
    arc_t *arc;

    arc = &pt->header->root[pt->header->arc_count++];
    arc->addr = addr;

    //db_printf("WPROF: allocating arc for %x\n", addr);
    return arc;
}


/*
 * append arc to the list
 */
static arc_t *wprof_arc_list_append(arc_t *root, arc_t *arc)
{
    arc_t *tmp;

    if(!root)
	return arc;

    tmp = root;
    while(tmp->next)
	tmp = tmp->next;
    tmp->next = arc;

    return root;
}


/*
 * find arc with address addr or allocate a new one
 */
static arc_t *wprof_find_arc(ptinfo_t *pt, arc_t *current, address32_t addr)
{
    arc_t *arc;

    for(arc = current->children; arc; arc = arc->next)
	{
	    if(arc->addr == addr)
		return arc;
	}

    arc = wprof_alloc_arc(pt, addr);
    current->children = wprof_arc_list_append(current->children, arc);

    return arc;
}

#define CHECKPT(pt)      do { if(!pt) return; if(pt == PTINFO_FINALIZED) return; } while(0)


/*
 * profile entry point
 * this is called first thing in a profiled function
 */
void __cyg_profile_func_enter(address32_t this_fn, address32_t call_site)
{
    ptinfo_t *pt;
    cstack_t *top;
    arc_t *arc;

    pt = wprof_get_ptinfo();
    CHECKPT(pt);

    if(pt->state == WPROF_STATE_OFF)
	    return;

    /* FIXME
     * what to do when a signal arrives in the middle of this?
     * for now we'll just ignore it and return
     */
    if(!compare_and_swap(&pt->state, WPROF_STATE_ON, WPROF_STATE_BUSY))
	{
	    db_printf("WPROF: warning; re-entry in __menter (state=%ld)\n", pt->state);
	    return;
	}

    top = &pt->cstack[pt->cstackp++];
    arc = wprof_find_arc(pt, top->current, this_fn);
    arc->count++;

    top++;
    top->current = arc;
    top->entry_time = get_tsc();
    
    compare_and_swap(&pt->state, WPROF_STATE_BUSY, WPROF_STATE_ON);
}


/*
 * called at exit of a profiled function
 */
void __cyg_profile_func_exit(address32_t this_fn, address32_t call_site)
{
    counter64_t after;
    ptinfo_t *pt;
    cstack_t *top;
    arc_t *arc;

    after = get_tsc();
    pt = wprof_get_ptinfo();
    CHECKPT(pt);

    if(pt->state == WPROF_STATE_OFF)
	    return;

    /*
     * this happens when __menter was re-entered (by a signal)
     * ignore the signal handler exit
     */
    if(pt->state == WPROF_STATE_BUSY)
	{
            db_printf("WPROF: warning; BUSY in __mexit\n");
	    return;
	}
    top = &pt->cstack[pt->cstackp];
    arc = top->current;
    arc->ticks += (after - top->entry_time);

    pt->cstackp--;

#if 0
    if(pt->cstackp <= 0) {
      db_printf("[%d] pt->cstackp = %d\n", WPROF_CURRENT, pt->cstackp);
      //wprof_finalize_thread(0);
    }
#endif
}


static void sim_func_exit(ptinfo_t *pt)
{
	counter64_t after;
	cstack_t *top;
	arc_t *arc;

	after = get_tsc();

	top = &pt->cstack[pt->cstackp];
	arc = top->current;
	arc->ticks += (after - top->entry_time);

	pt->cstackp--;
}



/*
 * grab a copy of /proc/self/maps and save the relevant entries to 
 * the profile output file.
 * FIXME: cant use /proc/self/.. since this may be called from a different process
 */
static void wprof_write_maps(ptinfo_t *pt, char *addr)
{
    FILE *fp;

    fp = fopen("/proc/self/maps", "r");
    if(!fp)
	return;
    
    while(!feof(fp))
	{
	    wprof_map_t map;
	    char tmp[1024], soname[256], perm[16];
	    int start, end, offset, res;

	    if(!fgets(tmp, 1024, fp))
		break;

	    res = sscanf(tmp, "%x-%x %s %x %*s %*d %[^\n]\n", 
			 &start, &end, perm, &offset, soname);
	    if(res != 5)
		continue;
	
	    /* only save entries mapped as executable */
	    if(!strstr(perm, "x"))
		continue;
	
	    memset(&map, 0, sizeof(wprof_map_t));
	    map.addr_from = start;
	    map.addr_to   = end;
	    map.offset    = offset;
	    strncpy(map.filename, soname, sizeof(map.filename));

	    memcpy(addr, &map, sizeof(wprof_map_t));
	    addr += sizeof(wprof_map_t);
	    pt->header->map_count++;
	}

    fclose(fp);
}


/*
 * finalize a threads output file
 */
static void wprof_finalize_thread(ptinfo_t *pt)
{
    size_t end;

    CHECKPT(pt);

    while(pt->cstackp > 0)
	{
	    db_printf("WPROF: unwind cstack %d\n", pt->cstackp);
	    sim_func_exit(pt);
	}

    pt->state = WPROF_STATE_OFF;
    
    db_printf("[%d] profile_exit\n", pt->pid);
    db_printf("[%d] arc_count=%d (mem=%d)\n",
	   pt->pid, pt->header->arc_count, 
	   pt->header->arc_count * sizeof(arc_t));
    
    end = sizeof(wprof_header_t) + (pt->header->arc_count * sizeof(arc_t));
    
    if(lseek(pt->map_fd, end, SEEK_SET) != -1) {
	    void *addr = (void *)pt->header;
	    int map_fd = pt->map_fd;
	    int map_size = pt->map_size;
	    
	    wprof_write_maps(pt, (char *)(pt->header->vaddr + end));
            //printf("[%d] Done...\n", pt->pid);
	    pt->header->flags |= WFLAG_CLOSED;
            wprof_unlink_ptinfo(pt);
	    
	    ftruncate(map_fd, end + (pt->header->map_count * sizeof(wprof_map_t)));
	    
	    munmap(addr, map_size);
	    close(map_fd);
    } else {
	    printf("WPROF: lseek failed\n");
    }
}


/*
 * exit the CURRENT thread
 */
void wprof_exit_thread(int status)
{
	ptinfo_t *pt;

	pt = wprof_get_ptinfo();
	CHECKPT(pt);

	/* record exit code */
	pt->header->exitcode = status;

	/* store some useful info */
	pt->header->flags |= WFLAG_HAS_RUSAGE;
	getrusage(RUSAGE_SELF, &pt->header->rusage);
	gettimeofday(&pt->header->time_stop, NULL);

	wprof_finalize_thread(pt);
}
