/* Wine Profiler
 *
 * Copyright 2000-2003 Codeweavers, Charles Loep
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>

#include <getopt.h>

#include <glib.h>

#include "profiler.h"
#include "wineprof.h"


/*--------------------------------------------------------*/
// options
/*--------------------------------------------------------*/
 
#define OPT_ON   1
#define OPT_OFF  0

struct options {
    int debug;
    int flat;
    int callers;
    int calltree;
    int soprefix;
    int merge;
    int listso;
    int source;
    char *sort;
    int precision;
    GList *ignore;
    int compress;
};

struct options opt = {
    debug     : OPT_OFF,
    flat      : OPT_ON,
    callers   : OPT_OFF,
    calltree  : OPT_OFF,
    soprefix  : OPT_OFF,
    merge     : OPT_OFF,
    listso    : OPT_OFF,
    source    : OPT_OFF,
    sort      : "self",
    precision : 3,
    ignore    : NULL,
    compress  : OPT_ON,
};


static const struct option long_options[] =
{
    { "flat",           no_argument,       &opt.flat,     OPT_ON  },
    { "no-flat",        no_argument,       &opt.flat,     OPT_OFF },
    { "callers",        no_argument,       &opt.callers,  OPT_ON },
    { "no-callers",     no_argument,       &opt.callers,  OPT_OFF },
    { "calltree",       no_argument,       &opt.calltree, OPT_ON  },
    { "no-calltree",    no_argument,       &opt.calltree, OPT_OFF },

    { "dso-prefix",     no_argument,       &opt.soprefix, OPT_ON  },
    { "no-dso-prefix",  no_argument,       &opt.soprefix, OPT_OFF },

    { "merge",          no_argument,       &opt.merge,    OPT_ON  },
    { "no-merge",       no_argument,       &opt.merge,    OPT_OFF },
    { "list-dso",       no_argument,       &opt.listso,   OPT_ON  },
    { "no-list-dso",    no_argument,       &opt.listso,   OPT_OFF },
    { "source",         no_argument,       &opt.source,   OPT_ON  },
    { "no-source",      no_argument,       &opt.source,   OPT_OFF },

    { "compress",       no_argument,       &opt.compress, OPT_ON  },
    { "no-compress",    no_argument,       &opt.compress, OPT_OFF },

    { "ignore",         required_argument, NULL,          'i' },
    { "sort",           required_argument, NULL,          's' },
    { "precision",      required_argument, NULL,          'p' },

    { "debug",          no_argument,       &opt.debug,    OPT_ON  },
    { "help",           no_argument,       NULL,          'h' },
};


/*--------------------------------------------------------*/
// prototypes
/*--------------------------------------------------------*/
 
void usage();
void view_profile_flat(profile_t *prof);
void view_profile_calltree(profile_t *prof);



/*--------------------------------------------------------*/
// the code!
/*--------------------------------------------------------*/


static unsigned long wprof_get_hz()
{
    FILE *fp;
    double ticks = 0.0;

    fp = fopen("/proc/cpuinfo", "r");
    while(!feof(fp))
        {

            char line[1024];

            if(!fgets(line, 1024, fp))
                break;
            if(sscanf(line, "cpu MHz         : %lf", &ticks) != 1)
                continue;
            printf("WPROF: ticks per second: %.0lf\n", (ticks * 1000000.0));
            break;
        }
    fclose(fp);

    if(ticks == 0.0)
        {
            printf("WPROF: unable to determine CPU HZ\n");
	    exit(1);
        }

    return (long)(ticks * 1000000.0);
}


static size_t file_size_for_fd(int fd)
{
    struct stat st;
    fstat(fd, &st);
    return st.st_size;
}

static char *basename(char *file)
{
    char *p = strrchr(file, '/');
    return p ? ++p : file;
}

static char *fullname(symbol_t *sym)
{
    static char tmp[1024];

    if(opt.soprefix)
	sprintf(tmp, "%s:%s", basename(sym->soname), sym->name);
    else
	return sym->name;

    return tmp;
}

static counter64_t msec(counter64_t hz, counter64_t count)
{
    double v;

    v = (double)count / ((double)hz / 1000.0);
    return (counter64_t)v;
}

static double dsec(counter64_t hz, counter64_t count)
{
    double v;

    v = (double)count / (double)hz;
    return v;
}

static double tv_to_double(struct timeval *tv)
{
	return (double)tv->tv_sec + ((double)tv->tv_usec / 1000000.0);
}


static char *repeat_char(char ch, int count)
{
	static char tmp[1024];
	memset(tmp, ch, count);
	tmp[count] = 0;
	return tmp;
}



/*
 * allocate a symbol
 */
symbol_t *symbol_new(char *name, address32_t addr)
{
    symbol_t *sym = g_new0(symbol_t, 1);
    sym->name = g_strdup(name);
    sym->addr = (unsigned int)addr;

    return sym;
}


/*
 * fetch all symbols from soname
 * <lazy> we use nm for now </lazy>
 */
static int read_symtab(GHashTable *hash, unsigned int base, char *soname)
{
    FILE *fp;
    char cmd[512], line[256];

    if(opt.source)
	sprintf(cmd, "nm --defined-only --line-numbers '%s' 2> /dev/null", soname);
    else
	sprintf(cmd, "nm --defined-only '%s' 2> /dev/null", soname);

    fp = popen(cmd, "r");
    if(!fp)
	return 0;

    soname = g_strdup(soname);

    while(!feof(fp))
	{
	    unsigned int addr = 0;
	    char type = 0;
	    char symbol[256], source[512] = { 0 };
	    symbol_t *sym;
	    int res = 0;

	    if(!fgets(line, 256, fp))
		break;

	    if(opt.source)
		res = sscanf(line, "%x %c %s %s\n", &addr, &type, symbol, source);
	
	    else if(!opt.source || res != 4)
		{
		    if(sscanf(line, "%x %c %s\n", &addr, &type, symbol) != 3)
			continue;
		}
	    
	    if(addr < base)
		addr += base;

	    //printf("%c %x %s\n", type, addr, symbol);

	    if(!g_hash_table_lookup(hash, (gpointer)addr) || type == 'T')
		{
		    sym = symbol_new(symbol, (address32_t)addr);
		    sym->soname = soname;
		    if(strlen(source))
			sym->source = g_strdup(source);
		    g_hash_table_insert(hash, (gpointer)addr, sym);
		}
	}
    pclose(fp);

    return 1;
}

/*
 * returns true if sym's name is in the ingore list
 */
static int symbol_is_ignored(symbol_t *sym)
{
    GList *e;

    if(!sym)
	return 0;
    //FIXME this should be a hashtable
    for(e = opt.ignore; e; e = e->next)
	{
	    char *str = e->data;
	    if(!strcmp(str, sym->name))
		return 1;
	}
    return 0;
}


/*
 * helper for symbol_add_caller
 */
static int cmp_callinfo_sym(gconstpointer a, gconstpointer b)
{
    callinfo_t *ci = (callinfo_t *)a;
    symbol_t *sym = (symbol_t *)b;

    return !(ci->caller == sym);
}

/*
 * add a caller to sym
 */
static void symbol_add_caller(symbol_t *sym, symbol_t *caller, arc_t *arc)
{
    callinfo_t *ci = 0;

    if(opt.compress)
	{
	    GList *e = g_list_find_custom(sym->callers, caller, cmp_callinfo_sym);
	    if(e)
		ci = e->data;
	}

    if(!ci)
	{
	    ci = g_new0(callinfo_t, 1);
	    ci->caller = caller;
	    sym->callers = g_list_append(sym->callers, ci);
	}

    ci->calls += arc->count;
    ci->ticks += arc->ticks;
}

counter64_t calc_all_ticks(arc_t *root, GHashTable *symtab, symbol_t *caller)
{
    counter64_t total_time = 0;
    counter64_t children_time = 0;
    arc_t *arc;

    if(!root)
        return 0;

    for(arc = root->children; arc; arc = arc->next)
        {
            symbol_t *sym;

            sym = g_hash_table_lookup(symtab, (gpointer)arc->addr);
	    if(!sym) { 
		    char name[32];
		    sprintf(name, "(0x%X)", (unsigned int)arc->addr);
		    sym = symbol_new(strdup(name), arc->addr);
		    //printf("SYMBOL NOT FOUND %x\n", arc->addr); 
                    g_hash_table_insert(symtab, (gpointer)arc->addr, sym);
	    }

            children_time = calc_all_ticks(arc, symtab, sym);
	    //printf("children_time = %llu\n", children_time);
            total_time += children_time;
	    if(!symbol_is_ignored(sym)) {
		    total_time += (arc->ticks - children_time);
		    
		    sym->count += arc->count;
                    sym->ticks += arc->ticks;
                    sym->ticks_kids += children_time;
                    sym->ticks_self += arc->ticks - children_time;

		    if(caller)
			{
			    symbol_add_caller(sym, caller, arc);
			    //printf("SYM %-20s  CALLER %s\n", sym->name, caller->name);
			}
		}
        }

    return total_time;
}


static GHashTable *process_maps(wprof_map_t *mp, int num_maps)
{
    GHashTable *hash;
    int i;

    hash = g_hash_table_new(g_direct_hash, g_direct_equal);

    for(i = 0; i < num_maps; i++)
        {
	    if(opt.listso)
		printf("%08x-%08x %08x %s  \n", mp->addr_from, mp->addr_to, mp->offset, mp->filename);
	    read_symtab(hash, mp->addr_from, mp->filename);
            mp++;
        }
    return hash;
}


profile_t *profile_open(char *filename)
{
    profile_t *prof;
    wprof_header_t header, *hp;
    double idle, systime, usertime;
    void *addr;
    int fd;

    fd = open(filename, O_RDONLY);
    read(fd, &header, sizeof(header));
    if(header.magic != WPROF_MAGIC)
        {
            printf("WPROF: invalid magic number in %s\n", filename);
            close(fd);
            return 0;
        }
    if(header.version != WPROF_VERSION)
	{
	    printf("WPROF: version mismatch in %s\n", filename);
	    close(fd);
	    return 0;
	}

    if(!(header.flags & WFLAG_CLOSED))
    {
	printf("WPROF: profile not properly closed.\n");
	//close(fd);
	//return 0;
      }

    prof = g_new0(profile_t, 1);

    usertime = tv_to_double(&header.rusage.ru_utime);
    systime  = tv_to_double(&header.rusage.ru_stime);

    idle = tv_to_double(&header.time_stop) - tv_to_double(&header.time_start);
    idle -= tv_to_double(&header.rusage.ru_utime) + tv_to_double(&header.rusage.ru_stime);
    
    if(opt.debug)
	{
	    printf("WPROF pid:   %d\n", header.pid);
	    printf("WPROF start: %s", ctime(&header.time_start.tv_sec));
	    printf("WPROF stop:  %s", ctime(&header.time_stop.tv_sec));
	    
	    printf("WPROF utime: %.3lf sec\n", usertime);
	    printf("WPROF stime: %.3lf sec\n", systime);
	    printf("WPROF idle:  %.3lf sec\n", idle);
	    
	    printf("WPROF exit:  %d\n", header.exitcode);
	    printf("WPROF clock: %lu\n", header.cpu_hz);
	    printf("WPROF addr:  %x\n", header.vaddr);
	    printf("WPROF #arc:  %d\n", header.arc_count);
	    printf("WPROF #map:  %d\n", header.map_count);
	}

    if(header.flags & WFLAG_HAS_RUSAGE) {
	    printf("pid   utime    stime    idle     total\n");
	    printf("----- -------- -------- -------- --------\n");
	    printf("%5d %8.3lf %8.3lf %8.3lf %8.3lf\n\n",
		   header.pid,
		   usertime,
		   systime,
		   idle,
		   usertime + systime + idle
		    );
    } else {
	    printf("WPROF: no rusage available.\n");
    }

    addr = mmap((void *)header.vaddr, file_size_for_fd(fd), PROT_READ,
                MAP_SHARED, fd, 0);
    if(addr == NULL || addr != (void *)header.vaddr)
	{
	    /* FIXME
	     * <lazy> we should relocate the file if it can't be mapped at the
	     * address specified in the header </lazy>
	     */
	    perror("mmap");
	    printf("Oops! failed to map profile at %x - bug Charles to fix it if this ever happens.\n", header.vaddr);
	    exit(1);
	}
    if(opt.debug)
	printf("WPROF mapped at: %p\n", addr);

    hp = (wprof_header_t *)addr;

    prof->fd     = fd;
    prof->header = hp;
    prof->symtab = process_maps((wprof_map_t *)(addr + sizeof(wprof_header_t) + (hp->arc_count * sizeof(arc_t))), hp->map_count);
    prof->total_ticks = calc_all_ticks(&hp->root[0], prof->symtab, 0);

    return prof;
}


void profile_close(profile_t *prof)
{
    //FIXME: free profile
}



static void do_make_list(gpointer key, gpointer value, gpointer user_data)
{
    GList *list = *(GList **)user_data;
    *(GList **)user_data = g_list_prepend(list, value);
}

GList *symtab_to_symlist(GHashTable *symtab)
{
    GList *list = 0;
    g_hash_table_foreach(symtab, do_make_list, &list);

    return list;
}



/*--------------------------------------------------------*/
// sorting functions
/*--------------------------------------------------------*/


#define COMPARE(a, b) ((a) == (b) ? 0 : (a) < (b) ? -1 : 1)  

#define DEFINE_SORT(TYPE, FIELD, NAME) \
gint NAME (gconstpointer a, gconstpointer b) \
{ \
  TYPE *val_a = (TYPE *)a; \
  TYPE *val_b = (TYPE *)b; \
  return COMPARE(val_b->FIELD, val_a->FIELD); \
}

DEFINE_SORT(symbol_t, count, do_sort_by_calls);
DEFINE_SORT(symbol_t, ticks, do_sort_by_ticks);
DEFINE_SORT(symbol_t, ticks_self, do_sort_by_ticks_self);
DEFINE_SORT(symbol_t, ticks_kids, do_sort_by_ticks_kids);
DEFINE_SORT(symbol_t, name, do_sort_by_name);


static GList *sort_symlist(GList *symlist)
{
    if(!strcmp(opt.sort, "name"))
	return g_list_sort(symlist, do_sort_by_name);

    if(!strcmp(opt.sort, "calls"))
        return g_list_sort(symlist, do_sort_by_calls);

    if(!strcmp(opt.sort, "ticks"))
        return g_list_sort(symlist, do_sort_by_ticks);

    if(!strcmp(opt.sort, "self"))
        return g_list_sort(symlist, do_sort_by_ticks_self);

    if(!strcmp(opt.sort, "child"))
        return g_list_sort(symlist, do_sort_by_ticks_kids);

    return symlist;
}

DEFINE_SORT(callinfo_t, caller->name, do_sort_ci_by_name);
DEFINE_SORT(callinfo_t, calls, do_sort_ci_by_calls);
DEFINE_SORT(callinfo_t, ticks, do_sort_ci_by_ticks);


static GList *sort_callinfo(GList *callers)
{
    if(!strcmp(opt.sort, "name"))
	return g_list_sort(callers, do_sort_ci_by_name);

    if(!strcmp(opt.sort, "calls"))
	return g_list_sort(callers, do_sort_ci_by_calls);
    
    return g_list_sort(callers, do_sort_ci_by_ticks);
}


/*
 * Display the flat profile
 */
void view_profile_flat(profile_t *prof)
{
    GList *list, *e;
    counter64_t total;
    counter64_t current = 0;
    counter64_t hz = prof->header->cpu_hz;

    list = symtab_to_symlist(prof->symtab);
    list = sort_symlist(list);

    total = prof->total_ticks;

    printf("\n\nFlat Profile:\n\n");

    if(opt.debug)
	{
	    printf("CPU HZ: %llu\n", hz);
	    printf("Total ticks: %llu  (%lf sec)\n", total, dsec(hz, total));
	}

    /* FIXME: add support for variable precision to other fields too */

    printf("\n");
    printf("%-*.*s cmul %% calls    self sec all sec  avg self avg all  function\n",
	   opt.precision + 4, opt.precision + 4, "% time");
    printf("%s ------ -------- -------- -------- -------- -------- --------\n", 
	   repeat_char('-', opt.precision + 4));

    for(e = list; e; e = e->next)
	{
	    symbol_t *sym = e->data;
	    double percent, cmul;
	    double sec_self, sec_all, sec_self_avg, sec_all_avg;

	    if(sym->count == 0)
		continue;
	    if(symbol_is_ignored(sym))
		continue;
	    //printf("sym->ticks_self = %lf  total = %lf\n", (double)sym->ticks_self, (double)total);
	    percent = (double)sym->ticks_self / (double)total * 100.0;

	    current += sym->ticks_self;
	    cmul = (double)current / (double)total * 100.0;

	    sec_self = dsec(hz, sym->ticks_self);
	    sec_all  = dsec(hz, sym->ticks_self + sym->ticks_kids);
	    
	    sec_self_avg = dsec(hz, sym->ticks_self / sym->count);
	    sec_all_avg  = dsec(hz, (sym->ticks_self + sym->ticks_kids) / sym->count);


	    printf("%*.*lf %6.2lf %8lu %8.3lf %8.3lf %8.3lf %8.3lf %s\n",
		   opt.precision + 4, opt.precision, percent,
		   cmul,
		   sym->count,
		   sec_self,
		   sec_all,
		   sec_self_avg,
		   sec_all_avg,
		   fullname(sym));

	    if(opt.source && sym->source)
		printf("- %s\n", sym->source);

#if 0
	    if(sym->callers)
		{
		    GList *ce;
		    int line = 0;
		    for(ce = sym->callers; ce; ce = ce->next)
			{
			    callinfo_t *ci = ce->data;
			    printf("%*s %8lu %s\n",
				   opt.precision+4+6+1, "",
				   ci->calls, 
				   ci->caller->name);
			    line++;
			}
		}
#endif

	}
}

/*
 * Display the callers profile
 */
void view_profile_callers(profile_t *prof)
{
    GList *list, *e;
    counter64_t total;
    counter64_t hz = wprof_get_hz(); //prof->header->cpu_hz;

    list = symtab_to_symlist(prof->symtab);
    list = sort_symlist(list);

    total = prof->total_ticks;

    printf("\n\nCallers Profile:\n\n");

    for(e = list; e; e = e->next)
	{
            symbol_t *sym = e->data;

	    if(sym->count == 0)
		continue;
	    
	    printf("%s  (%lu call%s, %.3lf sec) called by:\n", 
		   fullname(sym),
		   sym->count,
		   sym->count == 1 ? "" : "s",
		   dsec(hz, sym->ticks_kids + sym->ticks_self)
		   );
	    
	                                                          
	    printf("-calls----------- -time-------- -caller-------------\n");
	    if(sym->callers)
		{
		    GList *ce;
		    for(ce = sort_callinfo(sym->callers); ce; ce = ce->next)
			{
			    callinfo_t *ci = ce->data;
			    symbol_t *caller = ci->caller;

#if 1
			    printf("    %8lu %3.0f%% %8.3lf %3.0f%% %s\n",
				   ci->calls,
				   100.0 * ci->calls / sym->count,
				   dsec(hz, ci->ticks),
				   100.0 * ci->ticks / (sym->ticks_kids + sym->ticks_self),
				   caller->name);
#else
			    printf("    %s for %lu calls, %.3lf sec\n",
				   fullname(caller),
				   ci->calls,
				   dsec(hz, ci->ticks)
				   );
#endif
			}
		}
	    else
		printf("                                <unknown>\n");

	    printf("\n");

	}
}


static void do_view_profile_calltree(arc_t *root, GHashTable *symtab, int idx)
{
	arc_t *arc;

	if(!root)
		return;

	for(arc = root->children; arc; arc = arc->next)
        {
		symbol_t *sym;
		int i;

		sym = g_hash_table_lookup(symtab, (gpointer)arc->addr);

		for(i = 0; i < idx; i++)
			printf("  ");

		if(sym)
			printf("%s  (%lu calls, %llu ticks)\n", fullname(sym), arc->count, arc->ticks);
		else
			printf("%x  (%lu calls, %llu ticks)\n", (unsigned int)arc->addr, arc->count, arc->ticks);

		do_view_profile_calltree(arc, symtab, idx + 1);
        }
}

/*
 * Display the calltree
 */
void view_profile_calltree(profile_t *prof)
{
        printf("\n\nCalltree:\n\n");
        do_view_profile_calltree(&prof->header->root[0], prof->symtab, 0);
}




/*--------------------------------------------------------*/
// MAIN
/*--------------------------------------------------------*/


int main(int argc, char **argv)
{
    int option_index = 0;

    while(1)
	{
	    int c = getopt_long (argc, argv, "h?i:s:",
				 long_options, &option_index);
	    if(c == -1)
		break;

	    switch(c)
		{
		case 0:
		    break;

		    /*FIXME not sure if this option is useful/works correctly */
		case 'i':
		    {
			gchar **strv = g_strsplit(optarg, ",", 0);
			int i = 0;
			while(strv[i])
			    {
				opt.ignore = g_list_append(opt.ignore, g_strdup(strv[i]));
				i++;
			    }
			g_strfreev(strv);
		    }
		    break;

		case 's':
		    opt.sort = strdup(optarg);
		    break;
		    
		case 'p':
		    opt.precision = atoi(optarg);
		    break;

		case '?':
		case 'h':
		    usage();
		    exit(0);
		    break;
		default:
		    printf("Something is messed up ...\n");
		    break;
		}
	}


    if(optind >= argc)
	{
	    usage();
	    printf("\nError: no files specified.\n");
	    exit(1);
	}

    while(optind < argc)
	{
	    profile_t *prof;
	    char *filename = argv[optind++];

	    prof = profile_open(filename);
	    if(!prof)
		continue;

	    if(opt.flat)
		view_profile_flat(prof);

	    if(opt.callers)
		view_profile_callers(prof);

	    if(opt.calltree)
		view_profile_calltree(prof);

	    profile_close(prof);
	}

    exit(0);
}


void usage()
{
	static const char usage_msg[] = "Usage: wprof [OPTION]... [FILE]...\n\n"
		"  --flat            display flat profile\n"
		"  --callers         display callers profile\n"
		"  --calltree        display entire calltree (can be huge!)\n"
		"! --merge           merge specified profile files\n"
		"  --list-dso        list shared objects used by the program\n"
		"  --dso-prefix      prefix symbols with the name of the object they reside in\n"
		"  --ignore SYM      ignore symbol SYM (comma separated list)\n"
		"  --source          display source files and line numbers\n"
		"  --compress        compress callers with the same name into a single entry\n"
		"  --sort KEY        sort output by KEY (name, calls, ticks, self, child)\n"
		"  --help            display this help text\n\n"
		"Most options can be prefixed with 'no' to disable it, e.g. --no-flat\n";
	
	printf("%s", usage_msg);
}
